//--------------------------------------------------------------
// Copyright (C) 2003 Patrick Schaumont (schaum@ee.ucla.edu)                   
// University of California at Los Angeles
//                                                                             
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//                                                                             
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//                                                                             
// You should have received a copy of the GNU Lesser General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA
//
// $Id: armcosim.cxx,v 1.8 2005/01/19 02:54:27 schaum Exp $
//----------------------------------------------------------------

#include "armcosim.h"

static void usage(char *fname) {
  char *cp;
  if ((cp = strrchr(fname, '/'))!=NULL) cp++;
  else cp = fname;
  fprintf(stderr,
	  "usage : %s [-v] [-c max_cycles] [-f fpe-path] [-h] -g gezel_description -a arm_executable [arm_program_parameters]\n"
	  "  -v : SimIt-ARM verbose\n"
          "  -d : GEZEL value-change debug\n" 
	  "  -c : maximum number of clock cycles to simulate\n"
	  "  -f : FPE binary file path\n"
          "  -a : statically-linked arm-linux executable\n"
          "  -g : gezel hardware description\n"
	  "  -h : print this message and quit\n" 
          "ARM ISS by Wei Qin, visit http://sourceforge.net/projects/simit-arm/\n", cp);
}

/* ---------------------------------------------------- */

map <int, char *>  gezel_name;
map <int, char *>  processor_name;
map <int, char *>  exec_name;
map <int, char *>  args_name;
map <int, char **> processor_argv;
map <int, int>     processor_argc;

using namespace simulator;

rtsimgen *glbRTsimgen = 0;
map<int, arm_simulator *> glbSims; // SimIT-ARM-2.0

/* ---------------------------------------------------- */

static void sigint_handler(int signum) {
  for (map<int, arm_simulator *>::iterator i = glbSims.begin(); // SimIT-ARM-2.0
       i != glbSims.end();
       i++) {
    i->second->stop(); // Simit-ARM-2.0
  }
}

/* ---------------------------------------------------- */

void init_simulators(bool verbose, char *fpe_name, char *envp[]) {

  map<int, char *>::iterator i;


  for (i= processor_name.begin(); i != processor_name.end(); i++) {
    printf("Initializing StrongARM processor %s\n", i->second);

    arm_simulator *p = new arm_simulator(verbose); // Simit-arm-2.0
    p->load_fpe(fpe_name);
    p->load_program(exec_name[i->first], 
		    processor_argc[i->first], 
		    processor_argv[i->first], envp);

    p->initialize(); // Simit-ARM-2.0

    glbSims[i->first] = p;
  }

  if (gezel_name.size() > 0) {
    printf("Initializing GEZEL processor %s\n", gezel_name.begin()->second);

    /* load FDL program */
    fdldebug = 0;
    if (call_parser(gezel_name.begin()->second)) {
      fprintf(stderr, "Could not open %s\n", gezel_name.begin()->second);
      exit(0);
    }


    if (glbRTDebug) {
      ofstream F("fdl.symbols");
      glbSymboltable.show(F);
      F.close();
    }

    glbAllowSimulatorSleep = 1;
    glbMPZCustomAllocator = 0;
    glbRTsimgen = new rtsimgen;
    glbSymboltable.createcode(glbRTsimgen);
    //    mpzcl_init();
    glbRTsimgen->reset();
    
  }

}

/* ---------------------------------------------------- */

void run_simulators(UInt64 max_cnum) {
  UInt64 i;
  bool more = false;

  map <int, simulator::arm_simulator *>::iterator proc;  // Simit-arm-1.1

  for (i=0; (i<max_cnum); i++) {    
    more = false;
    for (proc = glbSims.begin(); 
	 proc != glbSims.end(); proc++) {
      if (proc->second->is_running()) { // simit-arm-2.0
	more |= true;
	proc->second->clock_tick();
      }
    }
    if (glbRTsimgen)
      glbRTsimgen->step();
    if (! more)
      break;
  }

  for (proc = glbSims.begin(); proc != glbSims.end(); proc++) 
  proc->second->dump_stats(stdout); // simit-arm-2.0

}

/* ---------------------------------------------------- */

int main(int argc, char *argv[], char *envp[]) {
  int i;
  char * prog_name = NULL;
  char * fpe_name  = NULL;
  bool   verbose   = false;
  UInt64 max_cnum  = (UInt64)-1;
  void (*prev_sigint_handler)(int);
  
  for(i = 1; i < argc; i++) {
    if(strcmp(argv[i], "-v") == 0) verbose = true; else
      if (strcmp(argv[i], "-d") == 0) glbRTDebug = 1; else
	if(strcmp(argv[i], "-f") == 0) fpe_name = argv[++i]; else
	  if(strcmp(argv[i], "-c") == 0) max_cnum = ato_uint64(argv[++i]); else
	    if(strcmp(argv[i], "-h") == 0) {usage(argv[0]); return 0;} else
	      if(strcmp(argv[i], "-g") == 0) gezel_name[0] = argv[++i]; else
		if(strcmp(argv[i], "-a") == 0) {
		  prog_name = argv[++i];
		  break;
		}
  }

  if (argc < 2) {
    usage(argv[0]);
    exit(0);
  }

  if (prog_name != NULL) {
    exec_name[0]      = prog_name;
    processor_argv[0] = argv + i;
    processor_argc[0] = argc - i;
  } else {
    usage(argv[0]);
    return 0;
  }

  if (gezel_name.size() == 0) {
    usage(argv[0]);
    return 0;
  }
  
  processor_name[0] = "arm_processor_0";
  init_simulators(verbose, fpe_name, envp);
  prev_sigint_handler = signal(SIGINT, sigint_handler);
  run_simulators(max_cnum);
  if (prev_sigint_handler != SIG_ERR) signal(SIGINT, SIG_DFL);
  
  return 0;
}




