//--------------------------------------------------------------
// Copyright (C) 2003 Patrick Schaumont (schaum@ee.ucla.edu)                   
// University of California at Los Angeles
//                                                                             
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//                                                                             
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//                                                                             
// You should have received a copy of the GNU Lesser General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA
//
// $Id: arm_itf.cxx,v 1.4 2004/07/21 23:19:37 schaum Exp $
//--------------------------------------------------------------
#define COSIM_STUB

#include "armzilla.h" // holds list of simulators
#include "arm_itf.h"
#include "external_mem.h"
#include "emumem.h"
#include "sysconparse.h" // holds list of processor names

// arm simulators are in this structure
// extern map<int, simulator::arm_simulator *> glbSims;
// processor names are in this structure (same index)
// extern map <int, char *>  processor_name;
//
// local structures
// map <int, aipblock *>      devicemap;
//
// (a) when parsing processor = "" name, 
//     look up name in processor_name.
//     find index, use index to retrieve current arm_simulator *
//     !! => this requires arm processors to be instantiated first
//     store this in the hook field
//
//
// (b) when parsing address = ""
//     call hook->emu.memory.register_addr(addr);
//     program integer returned into devicemap;
//     !! => if address is put before processor, 
//     current arm_simulator holds wrong value.
//     Therefore, order for processor/address is enforced
//     by armzilla_source & sink by means of the hook field
//
// ---- 
//
// portmap is always 0 for scalar hw/sw interfaces
// portmap is used for shared memory access (see below)

map <int, aipblock *>    devicemap; 
map <int, int>           portmap;
vector <aipblock *>      shmemlist;

// ---- ARMZILLASINK --- (memory-mapped HWSW cosimulation)

armzillasink::armzillasink(char *name) : aipblock(name) {
  hook = 0;
}

void armzillasink::setparm(char *_name) {
  gval *v = make_gval(32,0);
  char *proc;
  
  if ((proc = matchparm(_name, "process"))) {
    cerr << "armzillasink: set processor " << proc << "\n";
    int idx = find_proc_index(proc);
    if (idx < 0) {
      cerr << "armzillasink: processor " << proc << " does not exist\n";
      exit(0);
    }
    hook = glbSims[idx];
    return;
  }
  if (matchparm(_name, "address", *v)) {
    if (! hook) {
      cerr << "armzillasink: use address parm only after selecting a processor\n";
      exit(0);
    }
    cerr << "armzillasink: set address " << v->toulong() << "\n";

    unsigned deviceid = hook->mem->register_addr(v->toulong()); // simit-arm-2.0
    devicemap[deviceid] = this;
    portmap[deviceid] = 0;
    return;
  }
  cerr << "armzillasink: parameter " << _name << " not understood\n";
}

void armzillasink::run() {
  //  cout << "armzillasink value " << ioval[0].toulong() << "\n";
}

bool armzillasink::checkterminal(int n, char *tname, aipblock::iodir d) {
  switch (n) {
  case 0:
    return (isinput(d) && isname(tname, "data"));
    break;
  }
  return false;
}

bool armzillasink::cannotSleepTest() {
  return false;
}

// ---- ARMZILLASOURCE --- (memory-mapped HWSW cosimulation)

armzillasource::armzillasource(char *name) : 
  aipblock(name), 
  interfacewritten(false) {
  hook = 0;
}

void armzillasource::setparm(char *_name) {
  gval *v = make_gval(32,0);
  char *proc;

  if ((proc = matchparm(_name, "process"))) {
    cerr << "armzillasource: set processor " << proc << "\n";
    int idx = find_proc_index(proc);
    if (idx < 0) {
      cerr << "armzillasource: processor " << proc << " does not exist\n";
      exit(0);
    }
    hook = glbSims[idx];
    return;
  }

  if (matchparm(_name, "address", *v)) {
    if (! hook) {
      cerr << "armzillasource: use address parm only after selecting a processor\n";
      exit(0);
    }
    cerr << "armzillasource: set address " << v->toulong() << "\n";

    unsigned deviceid = hook->mem->register_addr(v->toulong()); // simit-arm-2.0
    devicemap[deviceid] = this;
    portmap[deviceid] = 0;
  }
}

void armzillasource::run() {
  //  cout << "armzillasink value " << ioval[0].toulong() << "\n";
}

bool armzillasource::checkterminal(int n, char *tname, aipblock::iodir d) {
  switch (n) {
  case 0:
    return (isoutput(d) && isname(tname, "data"));
    break;
  }
  return false;
}

bool armzillasource::needsWakeupTest() {
  bool v = interfacewritten;
  interfacewritten = false;
  return v;
}

bool armzillasource::cannotSleepTest() {
  bool v = interfacewritten;
  interfacewritten = false;
  return v;
}

void armzillasource::touch() {
  interfacewritten = true;
  //  printf("Touch armzillasource\n");
}

//---------------------------------------------------------------------
// shared memory

// principle:
// 
// the shared memory block contains a number of shared memory spaces for
// arm processors. In implementation, this memory corresponds to a multi-port
// memory unless additional access protocols can guarantee single-port access is
// feasible.

// the shared memory is created as a set of ioval[] values
// even though for gezel, the shared memory has no inputs or outputs.

// gezel definition example:
//
// ipblock shmem {
//   iptype "shmem";
//   ipparm "wlen=8";                  //  8 bits wordlength
//   ipparm "size=16";                 // 16 memory locations
//   ipparm "access=myproc@0x400000";  // arm 'myproc' sees this memory from 0x400000 to 0x40000F
//   ipparm "access=yourproc@0x200";   // arm 'yourproc' sees this memory from 0x200 to 0x20F
// }
//

//-------------------------------------------------------------------------------
// gezel port on shared memory
//
// ipblock shmem_writeport(in address : ns(5); in en : ns(1); in idata : ns(8)) {
//   iptype "shmem_writeport";
//   ipparm "shmem=shmem_ipblockname";
// }
//
// ipblock shmem_readport(in address : ns(5); in en : ns(1); out odata : ns(8)) {
//   iptype "shmem_readport";
//   ipparm "shmem=shmem_ipblockname";
// }

armzillashmem::armzillashmem(char *name) : aipblock(name) {
  size = 0;
  wl = 0;
  firstport = 0;
  shmemlist.push_back(this);
}

void armzillashmem::setparm(char *_name) {
  gval *v = make_gval(32,0);
  char *n;

  if (matchparm(_name, "size", *v)) {
    unsigned i;
    if (wl == 0) {
      cerr << "armzillashmem: wordlength parameter wl needs to be defined before defining size parameter\n";
      exit(0);
    }
    size = v->toulong();

    firstport = ioval.size(); // remember index of first port, to enable multiple shared memories

    for (i=0; i<size; i++)
      ioval.push_back(make_gval(wl,0)); // unsigned
  }

  if (matchparm(_name, "wl", *v)) {
    wl = v->toulong();
  }

  if ((n = matchparm(_name, "access"))) {
    // format: processor_name@address
    char *adr = n;
    while (*adr && (*adr != '@'))
      adr++;

    if (*adr == '@') {
      *adr = 0; // terminate processor name
      *adr++;   // address starts here
    } else {
      cerr << "armzillashmem: did not regnize access parameter " << n << "\n";
      cerr << "armzillashmem: access parameter format is <processorname>@<address>\n";
      exit(0);
    }

    int idx = find_proc_index(n);
    if (idx < 0) {
      cerr << "armzillashmem: processor " << n << " does not exist\n";
      exit(0);
    }

    unsigned adrval = make_gval(adr)->toulong();

    // define shared memory access
    unsigned i;
    for (i=0; i<size; i++) {
      // we multiply i by 4 to have 32-bit alignment of addresses
      unsigned deviceid = glbSims[idx]->mem->register_addr(adrval + 4*i); // simit-arm-2.0
      devicemap[deviceid] = this;
      portmap[deviceid] = i + firstport;
    }
  }
}

void armzillashmem::run() {
}

bool armzillashmem::cannotSleepTest() {
  return false;
}

void armzillashmem::touch() {
  //  unsigned i;
  //  for (i=0; i<size; i++) 
  //    cerr << i << " " << ioval[i]->toulong() << " ";
  //  cerr << "\n";
}

//------------------------------------------------------------

armzillashmem_writeport::armzillashmem_writeport(char *inst) : aipblock(inst), shmem(0) {}

void armzillashmem_writeport::setparm(char *_name) {
  char *n;

  if ((n = matchparm(_name, "shmem"))) {
    // find name n in shmemlist
    vector<aipblock *>::iterator it;
    for (it = shmemlist.begin(); it != shmemlist.end(); it++) {
      if ((*it)->blockisname(n)) { // found
	shmem = *it;
      }
    }
    if (shmem == 0) {
      cerr << "armzillashmem_writeport: shmem " << n << " must be defined before this port\n";
    }
  }
}

bool armzillashmem_writeport::checkterminal(int n, char *tname, aipblock::iodir dir) {
    switch (n) {
  case 0:
    return (isinput(dir) && isname(tname, "address"));
    break;
  case 1:
    return (isinput(dir) && isname(tname, "en"));
    break;
  case 2:
    return (isinput(dir) && isname(tname, "idata"));
    break;
  }
  return false;
}

void armzillashmem_writeport::run() {
  unsigned long value;
  if (shmem) {
    if (ioval[1]->toulong() != 0) {
      if (ioval[0]->toulong() >= shmem->ioval.size()) {
	cerr << "armzillashmem_writeport: address out of range\n";
      }
      value = ioval[2]->toulong();
      shmem->ioval[ ioval[0]->toulong() % shmem->ioval.size() ]->assignulong(value);
      shmem->touch();
    }
  }
}

bool armzillashmem_writeport::cannotSleepTest() {
  return false;
}

//------------------------------------------------------------

armzillashmem_readport::armzillashmem_readport(char *inst) : aipblock(inst), shmem(0) {}

void armzillashmem_readport::setparm(char *_name) {
  char *n;

  if ((n = matchparm(_name, "shmem"))) {
    // find name n in shmemlist
    vector<aipblock *>::iterator it;
    for (it = shmemlist.begin(); it != shmemlist.end(); it++) {
      if ((*it)->blockisname(n)) { // found
	shmem = *it;
      }
    }
    if (shmem == 0) {
      cerr << "armzillashmem_readport: shmem " << n << " must be defined before this port\n";
    }
  }
}

bool armzillashmem_readport::checkterminal(int n, char *tname, aipblock::iodir dir) {
  switch (n) {
  case 0:
    return (isinput(dir) && isname(tname, "address"));
    break;
  case 1:
    return (isinput(dir) && isname(tname, "en"));
    break;
  case 2:
    return (isoutput(dir) && isname(tname, "odata"));
    break;
  }
  return false;
}

void armzillashmem_readport::run() {
  if (shmem) {
    if (ioval[1]->toulong() != 0) {
      if (ioval[0]->toulong() >= shmem->ioval.size()) {
	cerr << "armzillashmem_readport: address out of range\n";
      }
      ioval[2]->assignulong(shmem->ioval[ ioval[0]->toulong() % shmem->ioval.size() ]->toulong());
      shmem->touch();
    }
  }
}

bool armzillashmem_readport::cannotSleepTest() {
  return false;
}

//----- MEMORY ACCESS INTERCEPT FUNCTIONS

byte_t     ext_read_byte     (int dev) {
  return devicemap[dev]->ioval[ portmap[dev] ]->toulong();
}

halfword_t ext_read_half_word(int dev) {
  return devicemap[dev]->ioval[ portmap[dev] ]->toulong();
}

word_t     ext_read_word     (int dev) {
  return devicemap[dev]->ioval[ portmap[dev] ]->toulong();
}

dword_t    ext_read_dword    (int dev) {
  return devicemap[dev]->ioval[ portmap[dev] ]->toulong();
}

void ext_write_byte     (int dev, byte_t value) {
  devicemap[dev]->ioval[ portmap[dev] ]->assignulong((long) value);
  devicemap[dev]->touch();
}

void ext_write_half_word(int dev, halfword_t value) {
  devicemap[dev]->ioval[ portmap[dev] ]->assignulong((long) value);
  devicemap[dev]->touch();
}

void ext_write_word     (int dev, word_t  value  ) {
  devicemap[dev]->ioval[ portmap[dev] ]->assignulong((long) value);
  devicemap[dev]->touch();
}

void ext_write_dword    (int dev, dword_t value  ) {
  devicemap[dev]->ioval[ portmap[dev] ]->assignulong((long) value);
  devicemap[dev]->touch();
}


