//--------------------------------------------------------------
// Copyright (C) 2003 Patrick Schaumont (schaum@ee.ucla.edu)                   
// University of California at Los Angeles
//                                                                             
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//                                                                             
// This program is distributed in the hope that it will be useful, but
// WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
// Lesser General Public License for more details.
//                                                                             
// You should have received a copy of the GNU Lesser General Public
// License along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
// USA
//
// $Id: ipblock.h,v 1.14 2005/01/19 01:04:54 herwin Exp $
//--------------------------------------------------------------
#ifndef IPBLOCK_H
#define IPBLOCK_H

#include "gval.h"
#include "rtsignal.h"
#include "rtio.h"
#include "rttrace.h"

//
// Operation of needsWakeupTest and cannotSleepTest
// After each run, cannotSleepTest is called and can return true/false
//    True: The block requires the next clock cycle to be simulated
//          because it either has an output that will change, or else
//          internal state that will change (like a counter).
//          By default, blocks can never sleep. The user must
//          explicitly overload cannotSleepTest and determine the
//          appropriate conditions under which false will be returned.
//    False: The block does not need simulation in the next clock cycle.
//           This is for example the case for purely combinatorial blocks
// When the simulator is in sleep mode, the simulator will call needsWakeupTest
// for each ip block in the simulation.
//    True: This clock cycle is a wakeup clock cycle, and the simulator must
//          transition to active mode.
//    False: This clock cycle is no wakeup cycle. By default, user blocks will
//          stay in sleep mode, and the needsWakeupTest needs to be explicitly
//          called and return true;
//

extern vector<aipblock *> glbIpblocks;

class aipblock {
  set <artsignal *>  inputs;
  set <artsignal *>  outputs;
  map <artsignal *, unsigned> iomap;
  long generation;
  bool ispresent;
  bool evalloop;
 public:
  vector<gval *> ioval;
 protected:
  char *name;
  enum iodir {input, output};
  bool isinput  (iodir d);
  bool isoutput (iodir d);
  bool isname   (char *, char *n);
  char * matchparm(char *orig, char *tgt);
  bool matchparm(char *orig, char *tgt, gval &v);
 public:
  aipblock(char *_name);
  virtual ~aipblock();
  void addoutput(artsignal *, char *terminal, int wlenreq, unsigned signreq);
  void addinput (artsignal *, char *terminal, int wlenreq, unsigned signreq);
  bool hasinput (artsignal  *);
  void eval(long g);  
  void present();
  gval &outval(artsignal *);
  bool blockisname(char *n);

  //---- user-def
  virtual void run();
  virtual void setparm(char *_name);
  virtual bool checkterminal(int n, char *tname, iodir dir);
  virtual bool needsWakeupTest();
  virtual bool cannotSleepTest();
  virtual void touch();
  virtual void probe(char *);
};

typedef rtinput ipblockin;

class ipblockout : public artsignal {
  aipblock *owner;
 public:
  ipblockout(symid _id, aipblock *owner);
  gval & eval(unsigned long g);
  gval &  peekval();
  int wlenreq();
  unsigned signreq();
};

//  #     #  #####  ####### ######
//  #     # #     # #       #     #
//  #     # #       #       #     #
//  #     #  #####  #####   ######
//  #     #       # #       #   #
//  #     # #     # #       #    #
//   #####   #####  ####### #     #

aipblock * ipblockcreate(char *instname, char *);

class tracer : public aipblock {
  tracerdumper *td;
 public:
  tracer(char *);
  ~tracer();
  void setparm(char *_name);
  void run();
  bool checkterminal(int n, char *tname, aipblock::iodir d);
  bool cannotSleepTest();
};

class filesource : public aipblock {
  ifstream *f;
  gval *rv;
  int  nfields;
 public:
  filesource(char *);
  ~filesource();
  void setparm(char *_name);
  void run();
  bool checkterminal(int n, char *tname, aipblock::iodir d);
  bool cannotSleepTest();
};

class ram : public aipblock {
  vector<gval *> content;
  int wl;

  ifstream *f;
  gval *rv;

  unsigned probestart;
  unsigned probestop;
 public:
  ram(char *);
  void setparm(char *_name);
  void run();
  bool checkterminal(int n, char *tname, aipblock::iodir d);
  bool cannotSleepTest();
  void probe(char *);
  void showcontents(unsigned from, unsigned to);
  void load_file();
};

class wideram : public aipblock {
  vector<gval *> content;
  int  wl;
  int  width;
 public:
  wideram(char *);
  void setparm(char *_name);
  void run();
  bool checkterminal(int n, char *tname, aipblock::iodir d);
  bool cannotSleepTest();
};

class gctr32 : public aipblock {
  gval *value, *one;
 public:
  gctr32(char *);
  void run();
  bool checkterminal(int n, char *tname, aipblock::iodir d);
};

class ctr32 : public aipblock {
  gval *value, *one;
 public:
  ctr32(char *);
  void run();
  bool checkterminal(int n, char *tname, aipblock::iodir d);
};

class gfmul16 : public aipblock {
  unsigned lasta, lastb;
 public:
  gfmul16(char *);
  void run();
  bool checkterminal(int n, char *tname, aipblock::iodir d);
};

class rand16 : public aipblock {
  unsigned lasta, lastb;
 public:
  rand16(char *);
  void run();
  bool checkterminal(int n, char *tname, aipblock::iodir d);
};

#endif

